'' Lander01 - Simple Moon Lander Simulation
''
'' Author: MSrobots
'' Version: 1
'' Date: 2018-09-07
''
'' Copyright (c) 2018 Michael Sommer
'' See end of file for terms of use.
''
CON                                 
                                                                              
  _clkmode = XTAL1|PLL16X                                                                    
  _xinfreq = 5_000_000

CON

  'Pin assignments
  VOL = 20       'volume level control pin
  SND = 21       'audio/sound output pin  
  SD_DO = 4      'SD Card DO/MISO
  SD_CK = 5      'SD Card Clock (SCLK)
  SD_DI = 9      'SD Card DI/MOSI  
  SD_CS = 15     'SD Card Chip Select, no pull-up resistor
  'data and clock pin assignments, respectively, in QuadKeyboard driver (in two places)
  'kbd0: 7, 8; kbd1: 12, 13; kbd2: 16, 17; kbd3: 23, 24  

  #0,  NORTH, WEST, SOUTH, EAST, NWSE     '4 scr nums (0..3), counterclockwise from north, plus all 4 screens (4)
  #10, QCN, QCW, QCS, QCE, QCC            '4 QC data for drawing quad chars: N, W, S, E and Colon
                                      
  #8,  BS, TAB, LF, VT, FF, CR, ESC = 27

  #0, MOUNT, UNMOUNT

  'foregnd-backgnd color pairs; avoid KK ($00) = Blk-on-Blk (string delimiter); don't use WR wht-on-red (reserved word)
  KK = $00, KR = $01, KG = $02, KB = $03, KY = $04, KP = $05, KA = $06, KW = $07 'blk foreground on some background (Note 1)
  RK = $10, RR = $11, RG = $12, RB = $13, RY = $14, RP = $15, RA = $16, RW = $17 'red foreground on some background
  GK = $20, GR = $21, GG = $22, GB = $23, GY = $24, GP = $25, GA = $26, GW = $27 'grn foreground on some backgroung  
  BK = $30, BR = $31, BG = $32, BB = $33, BY = $34, BP = $35, BA = $36, BW = $37 'blu foreground on some backgroung 
  YK = $40, YR = $41, YG = $42, YB = $43, YY = $44, YP = $45, YA = $46, YW = $47 'yel foreground on some backgroung
  PK = $50, PR = $51, PG = $52, PB = $53, PY = $54, PP = $55, PA = $56, PW = $57 'pur foreground on some backgroung     
  AK = $60, AR = $61, AG = $62, AB = $63, AY = $64, AP = $65, AA = $66, AW = $67 'aqu foreground on some backgroung  
  WK = $70, WS = $71, WG = $72, WB = $73, WY = $74, WP = $75, WA = $76, WW = $77 'wht foreground on some backgroung (Note 2)
  'Note 1: don't use KK ($00) = Blk-on-Blk in strings since 0 is the string delimiter value (for terminating strings) 
  'Note 2: for white-on-red, use white-on-scarlet (the scarlet letter) = WS because WR is a reserved word in PASM 
  'byte    %%0000, %%0010, %%0020, %%1000  ' 0: Black,      1: Red,        2: Green,      3: Blue  'Quaternary form
  'byte    %%0030, %%1010, %%1020, %%1030  ' 4: Yellow,     5: Purple,     6: Aqua        7: White 'Quaternary form

  gravity = 100 'The rate in which the spaceship descents in free fall (in ten seconds) 
  maxstep = 9 ' numbers of gridlines
  
OBJ

  vga[4] : "vga"            'VGA interface to VGA driver
  sdc    : "SD-MMC1A"       'SD card driver (FAT16), can also reload RAM from SD .bin file    
  kbd    : "QuadKeyboard"   'Quad keyboard driver (JRG)
''  rtc    : "DSRTC_driver2"  'RTC driver from Kwabena W. Agyeman (Kye), modified by Ron Czepala
VAR

  long state[4]                 'State of each Game
  long height[4]                'The height of the spaceship.
  long speed[4]                 'The speed of the spaceship.
  long burn[4]                  'The fuel which gets burned this step 
  long tensec[4]                'The time the flight is running for. (in ten second steps) 
  long fuel[4]                  'The fuel you have left. (kilogram) 
  long prevheight[4]            'The previous height to compare with actual. (for the colored digits) 
  long stepcount[4]             'Counts the steps passed since last output of the column names 

  long volume                   'speaker volume 0..20 (using long to avoid sign extend)   

  long rv                       'generic return value
  long sdErrMsg                 'SD card error string
  long sdErrNum                 'SD card error number
  byte sdProblem                'problem starting SD driver or mounting card; true/false
  
DAT '                                ###   THE MAIN LOOP   ###
PUB main | s, i, j, s0, s1, ch, today 
's=screen, i/j=gp, s0/s1=secs time 0/1, ch=char, rot=rotate, 'ch=kbd character

  initSystem     'start up system: video, keyboards, sd card, etc.

  splashScreens
  startGames
  
  repeat 'main loop, infinite
    s := (s + 1) // 4       'cycle among the keyboards (0..3, wrapping around)
    ch := kbd.getKey(s)            'get key in keyboard buffer, if present; 0 = no key
    if ch <> 0                       'check if there's a buffered keystroke for curr. kbd.
        
      if (state[s] < 1) and (ch <> 203)
        if state[s]<0
          startGame(s)
        else
          state[s] := 1           ' eat key - Start Game with input enabled
          drawScreen(s)           ' output new screen
          drawState(s)            ' output current values
      else  
        case ch
          8:
            truncateNumber(s)
          13:
            nextStep(s)
          48..58:
            addNumber(s,ch)
          200:
            truncateNumber(s)
          203:
            \sdc.bootPartition(string("/GAMEMENU/GAMEMENU.BIN")) 'launch GameMenu

DAT                           'Support Methods for MAIN LOOP Method        
PRI splashScreens | s, i 
  'show splash momentarily
  beep
  repeat s from 0 to 3
    clear(s, WW) 'clear screen to white
    repeat i from 0 to 4
      write(s, 3, i+2, GW, @BLand[i*34])
      write(s, 3, i+8, GW, @Ber01[i*34])
  waitcnt(clkfreq*5+cnt)
  beep

PRI startGames | s
  repeat s from 0 to 3
    startGame(s)

PRI startGame(s) | row
  state[s] := 0
  speed[s] := 1000
  height[s] := cnt & $FFFF
  ?height[s]
  height[s] := ((height[s] & $FFFF) // 15000) + 4000
  fuel[s] := 12000
  tensec[s] := 0
  burn[s] := 0
  prevheight[s] := height
  stepcount[s] := 1
  clear(s, WK) 'clear screen to black
  repChar(s, 0, 0, KG, 32, 40)
  write(s, 0, 0, KG,  @LunarLander)
  row := 2
  write(s, 0, row++, GK, string("This is a computer simulation of a"))
  write(s, 0, row++, GK, string("Apollo lunar landing capsule."))
  write(s, 0, row++, GK, string("The on-board computer has failed so you"))
  write(s, 0, row++, GK, string("have to land the capsule manually."))
  write(s, 0, row++, GK, string("Set burn rate of retro rockets to any"))
  write(s, 0, row++, GK, string("value between 0 (free fall) and"))
  write(s, 0, row++, GK, string("200 (maximum burn) kilo per second."))
  write(s, 0, row++, GK, string("Set burn rate for every 10 seconds."))
  write(s, 0, ++row, GK, string("Good Luck!"))
  repChar(s, 0, 14, KG, 32, 40)
  write(s, 0, 14, KG, @ExitStart)

PRI drawScreen(s)
  clear(s, WK) 'clear screen to black
  repChar(s, 0, 0, KG, 32, 40)
  write(s, 0, 0, KG, @LunarLander)
  write(s, 0, 2, GK, string(" Time  | Speed | Fuel  | Height| Burn   "))
  write(s, 0, 3, GK, string("-------+-------+-------+-------+--------"))
  repChar(s, 0, 14, KG, 32, 40)
  write(s, 0, 14, KG, string("ESC to EXIT - ENTER kg to burn:"))
  drawNumber6(s, 32, 14, RG, 0)
  
PRI drawState(s) | row
  row := stepcount[s]+3

  write(s, 0, row, GK, string("       |       |       |       |       "))
  drawNumber6(s, 0, row, GK, tensec[s])
  drawNumber6(s, 8, row, GK, speed[s])
  drawNumber6(s, 16, row, GK, fuel[s])
  drawNumber6(s, 24, row, GK, height[s])
  drawNumber6(s, 32, row, GK, burn[s])

PRI drawNumber6(s, c, r, k, n) | col, val, ch, sign
   val := n
   if val<0
     sign := "-"
     val := -val
   else
     sign := " "
   col := c + 5
   repeat 5
     ch := (val // 10) + 48
     val := val / 10
     putChar(s, col, r, k, ch)
     col--
   putChar(s, col, r, k, sign)

PRI addNumber(s, ch)
  burn[s] := (burn[s]*10)+ch-48
  if burn[s] > 200
    burn[s] := 200
  if burn[s]>fuel[s]
     burn[s] := fuel[s]
  drawNumber6(s, 32, 14, RG, burn[s])

PRI truncateNumber(s)
  burn[s] := burn[s]/10
  drawNumber6(s, 32, 14, RG, burn[s])

PRI nextStep(s)
  prevheight[s] := height[s]
  speed[s] += gravity - burn[s]                       ' calculate(height, speed, burn, gravity);
  height[s] -= speed[s]
  fuel[s] -= burn[s]
  tensec[s] += 10
  stepcount[s]++
  if stepcount[s] > maxstep
    stepcount[s] := 1
    drawScreen(s)
  drawState(s)
  burn[s] := 0
  drawNumber6(s, 32, 14, RG, burn[s])
  if height[s]<1
    if speed[s]>10
      repChar(s, 0, 13, KR, 32, 40)
      write(s, 0, 13, KR, string("YOU ARE DEAD!"))
    elseif (speed[s]<10 and speed[s]>3)
      repChar(s, 0, 13, GR, 32, 40)
      write(s, 0, 13, GR, string("YOU CRASHED!"))
    else
      repChar(s, 0, 13, RG, 32, 40)
      write(s, 0, 13, RG, string("SUCCESS! YOU LANDED AND SURVIVED!"))
    repChar(s, 0, 14, KG, 32, 40)   
    write(s, 0, 14, KG, @ExitStart)
    state[s] :=-1
    
DAT '                           OUTPUT FUNCTIONS FOR SCREENS
PUB setXYC(s, c, r, k)  'set cursor pos and color
  's = screen, c = column, r = row, k = color
  vga[s].out($A)
  vga[s].out(c)
  vga[s].out($B)
  vga[s].out(r)   
  vga[s].out($1B)         
  vga[s].out("c")
  vga[s].out(k)

PUB putChar(s, c, r, k, ch)  'put (not out) character
  's=screen, r=row, c=column, k=color, ch=character 
  setXYC(s, c, r, k)
  vga[s].putc(ch)

PUB putChar4(c, r, k, ch) | s  'put (not out) character to all screens
  's=screen, r=row, c=column, k=color, ch=character
  repeat s from 0 to 3
    setXYC(s, c, r, k)
    vga[s].putc(ch)
  
PUB write(s, c, r, k, str) 'output string to screen s at r, c in color k
  's=screen, c=column, r=row, k=color, str=string to write  
  vga[s].str2(c, r, k, str)

PUB write2(s, c, r, k, addr, n) 'output string to screen s at r, c in color k
  's=screen, c=column, r=row, k=color, str=string to write  
  vga[s].str3(c, r, k, addr, n)

PUB clear(s, k) | i, beg, end  'clear screen to solid background color
's = screen (0..3 or 4 for all screens), k = fore/background color
  beg := end := s
  if s => 4
    beg := NORTH
    end := EAST

  repeat i from beg to end
    vga[i].clear(k)

PUB repChar(s, c, r, k, ch, rep)  'output a single character severals times in a row
'display character ch from column c to column c+rep-1 (rep = repeat) on row r in color k
  setXYC(s, c, r, k)   
  repeat rep
    vga[s].putc(ch)

PUB showNN(s, c, r, k, val) | i, beg, end,d0,d1  'show a two-digit number on one or four screens
'draw two-digit number on one or all four screens
's=screen or use 4 for all, c=column, r=row, k=color, val=value to display
  beg := end := s
  if s == 4
    beg := 0
    end := 3

  d1 := val / 10
  d0 := val // 10
  repeat i from beg to end   
    putChar(i, c, r, k, d1+48)
    putChar(i, c+1, r, k, d0+48)

PUB showHeaderBar(s, k1, k2)
'show header bar on screen s in color k1 at top and middle and in k2 at bottom of header 
  repChar(s, 0, 0, k1, " ", 80)
  repChar(s, 0, 2, k2,  13, 40)  
    
DAT '                              ### SOUND/AUDIO ###
PUB tone(d, f)
'duration (number of wave cycles and relative frequency

  repeat d
    outa[SND]~                     'one wave cycle
    waitcnt(clkfreq/f+cnt)
    outa[SND]~~
    waitcnt(clkfreq/f+cnt) 

PUB beep
  tone(10, 1000)    'short beep, good for debugging
       
DAT '                                ### SD CARD ### 
PUB sdcall(funct)
  case funct
    0: sdc.mountPartition(0) ' Mount the default partition 0. Can be 0 - 3.
    1: sdc.unmountPartition

PUB startSDCard
  'start FAT driver for SD Card
  'do, clk, di, cs, wp, cd, rtcda, rtcclk, rtcbuslck/i2clock  
  rv := sdc.FATEngineStart(SD_DO, SD_CK, SD_DI, SD_CS, -1, -1, -1, -1, -1)
  sdProblem := !rv
  if !rv
    sdErrMsg := string("Driver could not start")

PUB restartSDC
'restart SD card driver and mount the first partition of the SD card
  \sdc.unmountPartition      'unmount the SD card (if mounted)
  sdc.FATEngineStop          'stop the SD card driver (if running)
  waitcnt(clkfreq/20+cnt)    'give SD card time, just in case
  startSDCard                'try starting the driver again
  mountSDCard                'and try mounting the SD card (if any)
    
PUB mountSDCard
  'mount sd card
  sdProblem := false
  sdErrMsg := \sdcall(MOUNT) 'Returns the address of the error string or null.
  sdErrNum := sdc.partitionError 'Returns the error number or zero.
  if sdErrNum > 0
    sdProblem := true

PUB stopSDCard
  sdErrMsg := \sdc.closeFile 'close the open file, if there is one open
  sdErrMsg := \sdcall(UNMOUNT)
  \sdc.FATEngineStop 

DAT '                          ### INITIALIZATION ###
PUB initSystem | s, i, vpins, lock

  'config audio pins
  dira[SND]~~                   'audio output pin
  dira[VOL]~~                   'volume control pin
  volume := 10                  'initial volume level  

  'set speaker volume
  ctra :=  %00110_000 << 23 + 1 << 9 + VOL   'counter mode = duty cycle for d/a using rc circuit (B pin is unused)
  frqa := $2000_0000                         'Vout = 3.3 * (frqa/2^32) = 3.3 * 0.5 = 1.65V max when frqa = $8000_0000 (can omit this line) 

  'configure video pins
  lock := ((cnt >> 16 + 2) | 1) << 16   
  repeat s from 0 to 3 
    vga[s].init(lock|vconfig[s])                'initialize vid drv with vgrp/vpins       

  repeat s from 0 to 3
    'vga[s].setn(1, @pal0)                       'switch to palette 0   
    vga[s].str(string($1B, "s"))                'page mode
    'vga[s].setn(1, constant(NEGX|$8000))        'use the ROM font for chars 0..31
    vga[s].setn(1, NEGX|@user_font)             'use the custom font for chars 0..31
    waitf(4) 'let voltage to screens stablize?

  'start quad keyboard driver
  kbd.start
 
  waitcnt(clkfreq/4+cnt)          'just in case
  startSDCard                     'start SDC driver
  waitcnt(clkfreq/4+cnt)          'just in case  
  mountSDCard                     'mount card, if present
    
  repeat i from 0 to 3            'defaults, etc.
    kbd.clearKeyBuffer(i)         'empty kbd buffers (probably not strictly necessary)
    
 
DAT '                                         UTILITY FUNCTIONS
PUB wait(sec)
  waitcnt(clkfreq*sec+cnt)

PUB waitf(div)
  waitcnt(clkfreq/div+cnt)
   
DAT 'vconfig
vconfig long    0 << 9 | %%103_3     'with sync   %010011_11  
        long    1 << 9 | %%103_0     'no sync     %010011_00      
        long    2 << 9 | %%103_0     '   
        long    3 << 9 | %%103_0     '           
LunarLander   byte "LUNAR LANDER 01",0
ExitStart     byte "ESC to EXIT - any other key to START    ",0

BLand         byte 32,32,06,14,32,32,32,32, 32, 06,14,14,14,14,14,07, 32, 06,14,07,32,32,14,07, 32, 14,14,14,14,14,14,07,32, 0 '34
              byte 32,32,14,14,32,32,32,32, 32, 14,14,32,32,32,14,14, 32, 14,14,14,07,32,14,14, 32, 14,14,32,32,32,14,14,32, 0 
              byte 32,32,14,14,32,32,32,32, 32, 14,14,14,14,14,14,14, 32, 14,14,09,14,07,14,14, 32, 14,14,32,32,32,14,14,32, 0
              byte 32,32,14,14,32,32,32,32, 32, 14,14,32,32,32,14,14, 32, 14,14,32,09,14,14,14, 32, 14,14,32,32,32,14,14,32, 0
              byte 32,32,09,14,14,14,14,08, 32, 09,14,32,32,32,14,08, 32, 09,14,32,32,09,14,08, 32, 14,14,14,14,14,14,08,32, 0  

Ber01         byte 32,32,06,14,14,14,14,07, 32, 06,14,14,14,14,14,07, 32, 32,06,14,14,14,14,07, 32, 32,32,14,14,32,32,32,32, 0 '34
              byte 32,32,14,14,32,32,32,32, 32, 14,14,32,32,32,14,14, 32, 32,14,14,32,06,14,14, 32, 32,32,14,14,32,32,32,32, 0
              byte 32,32,14,14,14,14,14,32, 32, 14,14,14,14,14,14,08, 32, 32,14,14,06,08,14,14, 32, 32,32,14,14,32,32,32,32, 0
              byte 32,32,14,14,32,32,32,32, 32, 14,14,32,09,14,07,32, 32, 32,14,14,08,32,14,14, 32, 32,32,14,14,32,32,32,32, 0
              byte 32,32,09,14,14,14,14,08, 32, 09,14,32,32,09,14,07, 32, 32,09,14,14,14,14,08, 32, 32,32,14,14,32,32,32,32, 0

DAT

{
  *** This technique is from Roger Williams' vga_40x18_rom_demo.spin v1.1 ***
  
  The user defined chars must be reversed and interleaved like the ROM font,
  .... This example shows a way to draw the font forward and un-interleaved
  so that the proptool will assemble it correctly.  Draw the left/lsb=0
  character with 2's and the right/lsb=1 character with 1's.  The reverse
}

user_font 'beginning address of user's custom font data

              'Cust Char 0: FIT-Blank     Cust Char 1: Vert-Split
C00_01  long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32  'Note: the slightly shortend underscore makes 
        long  %%00000000_00000000 >< 32 + %%11111111_00000000 >< 32  '...them more countable for filling in blanks

              'Cust Char 2: Banner        Cust Char 3: 3px Top Bar
C02_03  long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32

              'Cust Char 4: Left V-Bar    Cust Char 5: Right V-Bar        
C04_05  long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32                                                         
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32                 

              'Cust Char 6: Diag1         Cust Char 7: Diag2                in the order of an 'o' clk-wise from up-lf corner (6-9)
C06_07  long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32  
        long  %%00000000_00000002 >< 32 + %%10000000_00000000 >< 32
        long  %%00000000_00000002 >< 32 + %%10000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%11000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%11000000_00000000 >< 32
        long  %%00000000_00000222 >< 32 + %%11100000_00000000 >< 32
        long  %%00000000_00000222 >< 32 + %%11100000_00000000 >< 32                                                         
        long  %%00000000_00002222 >< 32 + %%11110000_00000000 >< 32
        long  %%00000000_00002222 >< 32 + %%11110000_00000000 >< 32
        long  %%00000000_00022222 >< 32 + %%11111000_00000000 >< 32
        long  %%00000000_00022222 >< 32 + %%11111000_00000000 >< 32
        long  %%00000000_00222222 >< 32 + %%11111100_00000000 >< 32
        long  %%00000000_00222222 >< 32 + %%11111100_00000000 >< 32
        long  %%00000000_02222222 >< 32 + %%11111110_00000000 >< 32
        long  %%00000000_02222222 >< 32 + %%11111110_00000000 >< 32
        long  %%00000000_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%00000000_22222222 >< 32 + %%11111111_00000000 >< 32
        long  %%00000002_22222222 >< 32 + %%11111111_10000000 >< 32
        long  %%00000002_22222222 >< 32 + %%11111111_10000000 >< 32
        long  %%00000022_22222222 >< 32 + %%11111111_11000000 >< 32
        long  %%00000022_22222222 >< 32 + %%11111111_11000000 >< 32
        long  %%00000222_22222222 >< 32 + %%11111111_11100000 >< 32
        long  %%00000222_22222222 >< 32 + %%11111111_11100000 >< 32
        long  %%00002222_22222222 >< 32 + %%11111111_11110000 >< 32
        long  %%00002222_22222222 >< 32 + %%11111111_11110000 >< 32
        long  %%00022222_22222222 >< 32 + %%11111111_11111000 >< 32
        long  %%00022222_22222222 >< 32 + %%11111111_11111000 >< 32
        long  %%00222222_22222222 >< 32 + %%11111111_11111100 >< 32
        long  %%00222222_22222222 >< 32 + %%11111111_11111100 >< 32
        long  %%02222222_22222222 >< 32 + %%11111111_11111110 >< 32
        long  %%02222222_22222222 >< 32 + %%11111111_11111110 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32          

              'Cust Char 8: Diag3         Cust Char 9: Diag4         
C08_09  long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32  
        long  %%22222222_22222220 >< 32 + %%01111111_11111111 >< 32
        long  %%22222222_22222220 >< 32 + %%01111111_11111111 >< 32
        long  %%22222222_22222200 >< 32 + %%00111111_11111111 >< 32
        long  %%22222222_22222200 >< 32 + %%00111111_11111111 >< 32
        long  %%22222222_22222000 >< 32 + %%00011111_11111111 >< 32
        long  %%22222222_22222000 >< 32 + %%00011111_11111111 >< 32                                                         
        long  %%22222222_22220000 >< 32 + %%00001111_11111111 >< 32
        long  %%22222222_22220000 >< 32 + %%00001111_11111111 >< 32
        long  %%22222222_22200000 >< 32 + %%00000111_11111111 >< 32
        long  %%22222222_22200000 >< 32 + %%00000111_11111111 >< 32
        long  %%22222222_22000000 >< 32 + %%00000011_11111111 >< 32
        long  %%22222222_22000000 >< 32 + %%00000011_11111111 >< 32
        long  %%22222222_20000000 >< 32 + %%00000001_11111111 >< 32
        long  %%22222222_20000000 >< 32 + %%00000001_11111111 >< 32
        long  %%22222222_00000000 >< 32 + %%00000000_11111111 >< 32
        long  %%22222222_00000000 >< 32 + %%00000000_11111111 >< 32
        long  %%22222220_00000000 >< 32 + %%00000000_01111111 >< 32
        long  %%22222220_00000000 >< 32 + %%00000000_01111111 >< 32
        long  %%22222200_00000000 >< 32 + %%00000000_00111111 >< 32
        long  %%22222200_00000000 >< 32 + %%00000000_00111111 >< 32
        long  %%22222000_00000000 >< 32 + %%00000000_00011111 >< 32
        long  %%22222000_00000000 >< 32 + %%00000000_00011111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22220000_00000000 >< 32 + %%00000000_00001111 >< 32
        long  %%22200000_00000000 >< 32 + %%00000000_00000111 >< 32
        long  %%22200000_00000000 >< 32 + %%00000000_00000111 >< 32
        long  %%22000000_00000000 >< 32 + %%00000000_00000011 >< 32
        long  %%22000000_00000000 >< 32 + %%00000000_00000011 >< 32
        long  %%20000000_00000000 >< 32 + %%00000000_00000001 >< 32
        long  %%20000000_00000000 >< 32 + %%00000000_00000001 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32 

              'Cust Char 10: bullet       Cust Char 11: Reg Blank         
C10_11  long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32  
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002200_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002220_00000000 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00002222_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_20000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222200 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222220 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222220 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222200 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_22000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_20000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00002222_00000000 >< 32 + %%01111111_11111100 >< 32
        long  %%00002220_00000000 >< 32 + %%01111111_11111100 >< 32
        long  %%00002200_00000000 >< 32 + %%01111111_11111100 >< 32
        long  %%00002000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32

              'Cust Char 12: Block         Cust Char 13: Low blank         
C12_13  long  %%22222222_22222222 >< 32 + %%01111111_11111100 >< 32  
        long  %%22222222_22222222 >< 32 + %%01111111_11111100 >< 32
        long  %%22222222_22222222 >< 32 + %%01111111_11111100 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32 

              'Cust Char 14: Block         Cust Char 15: Mesh         
C14_15  long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32  
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32                                                         
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32
        long  %%22222222_22222222 >< 32 + %%01010101_01010101 >< 32
        long  %%22222222_22222222 >< 32 + %%10101010_10101010 >< 32 

              'Cust Char 16: TBD          Cust Char 17: ellipsis...         
C16_17  long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32  
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000022 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000222 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000220 >< 32 + %%00000000_00000000 >< 32                                                         
        long  %%00000000_00002220 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00002220 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00002220 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00022200 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00022200 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00022200 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_02222000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_02220000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%20000000_22220000 >< 32 + %%00000000_00000000 >< 32
        long  %%22000002_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%02200002_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%02220022_22200000 >< 32 + %%00000000_00000000 >< 32
        long  %%00222222_22000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00222222_22000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00022222_22000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00022222_20000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00002222_20000000 >< 32 + %%00111001_11001110 >< 32
        long  %%00002222_20000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000222_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000222_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32
        long  %%00000000_00000000 >< 32 + %%00000000_00000000 >< 32

              'Cust Char X: Logo-TBar     Cust Char X: Logo-BBar   
C18_19  long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%00000000_00000000 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32                                                         
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%11111111_11111111 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32
        long  %%22222222_22222222 >< 32 + %%00000000_00000000 >< 32        

DAT
{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}             