''**********************************************************************************************************************
'' Quad PS/2 Keyboard Driver                                                                                           
'' (C) Jim Goodpaster                                                                                                 
'' Version KB7B works for 4 keyboards, but needs individual buffers
''
'' Update June 4th, 2016: not sure where I left off
'' Needs reset code (handled individually)                                                
''**********************************************************************************************************************
CON

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000

VAR

  long cog
  '11 continuous longs are used to implement the keyboard buffers for four keyboards
  byte  kb[32] 'four 8-byte buffers to buffer 7 keystrokes for 4 keyboards (one slot each is empty for management)
  byte  kh[4], kt[4] 'head and tail ptrs for 4 keyboards; keystrokes are input at the head

PUB start : okay | c, i 
  
' Starts quad keyboard driver for 4 keyboards
  stop
  okay := cog := cognew(@entry, @kb) + 1

' Intialize KB buffers
  repeat i from 0 to 31
    kb[i] := $00 'virgin byte

' Initialize head and tail pointers
  repeat i from 0 to 3
    kh[i] := kt[i] := 0 'empty buffer                    
  
PUB hasKey(kbn) 
  'if head = tail then buffer is empty
  if kbn > 4 or kbn < 0
    result := false
  else
    result := kh[kbn] <> kt[kbn] 

PUB clearKeyBuffer(kbn)
  'clear the keyboard buffer for given keyboard
  kh[kbn] := kt[kbn]

PUB getKey(kbn)
  'pulls character from buffer of current kb (kbn); chars are pulled from the head here (and added to the tail in pasm)
  result := 0 'default
  if kh[kbn] <> kt[kbn]
    result := kb[kbn*8+kh[kbn]]
    kh[kbn] := (kh[kbn] + 1) // 8

  'Note: to test code for full buffer, use something like waitcnt(clkfreq*10 + cnt)

PUB stop

'' Stop keyboard driver - frees a cog
                                                        
  if cog
    cogstop(cog~ -  1)

DAT 'quad keyboard driver
'http://www.computer-engineering.org/ps2protocol/

                        org     

'entry
entry                   mov     delay, d1sec            'wait one second to let kbd pwr-up and begin listening
                        call    #pause
                        
                        call    #reset_w                'reset W/0 keyboard
                        call    #reset_x                'reset X/1 keyboard
                        call    #reset_y                'reset Y/2 keyboard                         
                        call    #reset_z                'reset Z/3 keyboard

'#########################################################################################################################
'################################################ BEGIN MAIN LOOP ########################################################
'#########################################################################################################################

loop                    'main, outermost keyboard scanning loop 

'#########################################################################################################################
'#########################################$$### KEYBOARD 0/W = NORTH #####################################################
'#########################################################################################################################

kb_w                    mov     kbid, #0                'setup keyboard ID
                        mov     f8421, #1               'for key flags: 0001 for kb 0/W, 0010 for kb 1/X, etc.

                        mov     kbcw, ina               'read port and mask clock line
                        and     kbcw, kbcmw

                        cmp     kbcw, #0        wz      'check if clock is low
                        cmp     kbcpw, #1       wc      'check if data bit has already been read
        if_c_and_z      mov     kbcpw, #1               'mark start of low clock pulse (set if clk=z=0 & cp=c=0)
        if_c_and_z      mov     kbdw, ina               'read data line (data is valid 5us before falling edge of clk)
        if_c_and_z      and     kbdw, kbdmw             'mask off data bit
        if_c_and_z      shr     kbdbw, #1               'make way for new bit 
        if_c_and_z      shl     kbdw, #2                'move data bit to msb CHANGE LITERAL (eventually dumps start bit)
        if_c_and_z      or      kbdbw, kbdw             'prepend new bit to data bits (prepending reverses bit order)
        if_c_and_z      add     kbdbcw, #1              'increment the count for the number of received bits (0..11)
        
        if_nc_and_nz    mov     kbcpw, #0               'clock went high so reset clock pulse marker to prepare for next bit

                        cmp     kbdbcw, #11     wc      'check if 11 bits were received (though start bit gets ror'd out)
        if_c_or_z       jmp     #:cont                  'jump if either <11 bits or if clk line is low    

                        and     kbdbw, #$FF             'lop off parity and stop bits                              

        '############## at this point, a scancode has been received in kbdb* (from a pressed keyboard key)  ##############                

                        cmp     kbdbw, #$AA     wz      'check if keyboard needs reset
        if_z            call    #reset_w
        if_z            mov     delay, d1sec          
        if_z            call    #pause        
        if_z            jmp     #:skip    

        '############## process scan code from keyboard to convert it to ASCII character, including shift   ##############
          
                        mov     kbdb, kbdbw             'setup for filter call
                        call    #filter                 'preprocess scancode           
                        tjnz    jump, #:skip            'jump to skip if jump is set              

        '############## at this point, the scancode is not FO, EO nor the byte after those, and also not E1 ##############

                        call    #getchar                'get character and process shift keys (CHANGE?)
                        tjz     rcode, #:skip           'toss out any chars that mapped to 0 that slipped through filter

                        'write character to keyboard buffer area of shared hub memory
                        call    #write                  'write character to buffer (if not full)
        
:skip                   'peform clean up in prep for next byte from keyboard 
                        mov     kbcpw, #0               'reset clock pulse marker
                        mov     kbdbcw, #0              'reset counter for the number of received data bits to zero
                        mov     kbdbw, #0               'clear any residual bits/reset keyboard data bits (REMOVE?)
                        mov     kbtocw, #0              'reset timeout counter

                        'check for timeout in case of "hang" during a character due to line noise/glitch 
:cont                   tjz     kbdbcw, #kb_x           'check if any of the 11 data bits have been received 
                        add     kbtocw, #1              'if at least 1 bit has been received, increment time-out counter
                        cmp     kbtocw, timeout wc      'compare current count with the timeout constant value
        if_nc           jmp     #:skip                  'timeout occurred, reset for next key

'#########################################################################################################################
'################################################# KEYBOARD 1/X ##########################################################
'#########################################################################################################################

kb_x                    mov     kbid, #1                'setup keyboard ID
                        mov     f8421, #2               'for key flags: 0001 for kb 0/W, 0010 for kb 1/X, etc.
 
                        mov     kbcx, ina               'read port and mask clock line
                        and     kbcx, kbcmx

                        cmp     kbcx, #0        wz      'check if clock is low
                        cmp     kbcpx, #1       wc      'check if data bit has already been read
        if_c_and_z      mov     kbcpx, #1               'mark start of low clock pulse (set if clk=z=0 & cp=c=0)
        if_c_and_z      mov     kbdx, ina               'read data line (data is valid 5us before falling edge of clk)
        if_c_and_z      and     kbdx, kbdmx             'mask off data bit
        if_c_and_z      shr     kbdbx, #1               'make way for new bit            
        if_c_and_z      shr     kbdx, #3                'move data bit to msb CHANGE LITERAL (eventually dumps start bit)
        if_c_and_z      or      kbdbx, kbdx             'prepend new bit to data bits (prepending reverses bit order)
        if_c_and_z      add     kbdbcx, #1              'increment the count for the number of received bits (0..11)
        
        if_nc_and_nz    mov     kbcpx, #0               'clock went high so reset clock pulse marker to prepare for next bit

                        cmp     kbdbcx, #11     wc      'check if 11 bits were received (though start bit gets ror'd out)
        if_c_or_z       jmp     #:cont                  'jump if either <11 bits or if clk line is low    

                        and     kbdbx, #$FF             'lop off parity and stop bits

        '############## at this point, a scancode has been received in kbdb* (from a pressed keyboard key)  ##############                

                        cmp     kbdbx, #$AA     wz      'check if keyboard needs reset
        if_z            call    #reset_x
        if_z            mov     delay, d1sec          
        if_z            call    #pause        
        if_z            jmp     #:skip    

        '############## process scan code from keyboard to convert it to ASCII character, including shift   ##############
                
                        mov     kbdb, kbdbx             'setup for filter call
                        call    #filter                 'preprocess scancode
                        tjnz    jump, #:skip            'jump to skip if jump is set         

        '############## at this point, the scancode is not FO, EO nor the byte after those, and also not E1 ##############

                        call    #getchar                'get character and process shift keys (CHANGE?)
                        tjz     rcode, #:skip           'toss out any chars that mapped to 0 that slipped through filter                        
                        
                        'write character to keyboard buffer area of shared hub memory
                        call    #write                  'write character to buffer (if not full)
                        
:skip                   'peform clean up in prep for next byte from keyboard 
                        mov     kbcpx, #0               'reset clock pulse marker
                        mov     kbdbcx, #0              'reset counter for the number of received data bits to zero
                        mov     kbdbx, #0               'clear any residual bits/reset keyboard data bits (REMOVE?)
                        mov     kbtocx, #0              'reset timeout counter                         

                        'check for timeout in case of "hang" during a character due to line noise/glitch 
:cont                   tjz     kbdbcx, #kb_y           'check if any of the 11 data bits have been received 
                        add     kbtocx, #1              'if at least 1 bit has been received, increment time-out counter
                        cmp     kbtocx, timeout wc      'compare current count with the timeout constant value
        if_nc           jmp     #:skip                  'timeout occurred, reset for next key

'#########################################################################################################################
'################################################# KEYBOARD 2/Y ##########################################################
'#########################################################################################################################
 
kb_y                    mov     kbid, #2                'setup keyboard ID
                        mov     f8421, #4               'for key flags: 0001 for kb 0/W, 0010 for kb 1/X, etc.
                    
                        mov     kbcy, ina               'read port and mask clock line
                        and     kbcy, kbcmy

                        cmp     kbcy, #0        wz      'check if clock is low
                        cmp     kbcpy, #1       wc      'check if data bit has already been read
        if_c_and_z      mov     kbcpy, #1               'mark start of low clock pulse (set if clk=z=0 & cp=c=0)
        if_c_and_z      mov     kbdy, ina               'read data line (data is valid 5us before falling edge of clk)
        if_c_and_z      and     kbdy, kbdmy             'mask off data bit
        if_c_and_z      shr     kbdby, #1               'make way for new bit            
        if_c_and_z      shr     kbdy, #7                'move data bit to msb CHANGE LITERAL (eventually dumps start bit)
        if_c_and_z      or      kbdby, kbdy             'prepend new bit to data bits (prepending reverses bit order)
        if_c_and_z      add     kbdbcy, #1              'increment the count for the number of received bits (0..11)
        
        if_nc_and_nz    mov     kbcpy, #0               'clock went high so reset clock pulse marker to prepare for next bit

                        cmp     kbdbcy, #11     wc      'check if 11 bits were received (though start bit gets ror'd out)
        if_c_or_z       jmp     #:cont                  'jump if either <11 bits or if clk line is low    

                        and     kbdby, #$FF             'lop off parity and stop bits

        '############## at this point, a scancode has been received in kbdb* (from a pressed keyboard key)  ##############                

                        cmp     kbdby, #$AA     wz      'check if keyboard needs reset
        if_z            call    #reset_y
        if_z            mov     delay, d1sec          
        if_z            call    #pause        
        if_z            jmp     #:skip    

        '############## process scan code from keyboard to convert it to ASCII character, including shift   ##############

                        mov     kbdb, kbdby             'setup for filter call
                        call    #filter                 'preprocess scancode
                        tjnz    jump, #:skip            'jump to skip if jump is set             

        '############## at this point, the scancode is not FO, EO nor the byte after those, and also not E1 ##############

                        call    #getchar                'get character and process shift keys (CHANGE?)
                        tjz     rcode, #:skip           'toss out any chars that mapped to 0 that slipped through filter                        
                        
                        'write character to keyboard buffer area of shared hub memory
                        call    #write                  'write character to buffer (if not full)
                        
:skip                   'peform clean up in prep for next byte from keyboard 
                        mov     kbcpy, #0               'reset clock pulse marker
                        mov     kbdbcy, #0              'reset counter for the number of received data bits to zero
                        mov     kbdby, #0               'clear any residual bits/reset keyboard data bits (REMOVE?)
                        mov     kbtocy, #0              'reset timeout counter                         
                                                        
                        'check for timeout in case of "hang" during a character due to line noise/glitch 
:cont                   tjz     kbdbcy, #kb_z           'check if any of the 11 data bits have been received 
                        add     kbtocy, #1              'if at least 1 bit has been received, increment time-out counter
                        cmp     kbtocy, timeout wc      'compare current count with the timeout constant value
        if_nc           jmp     #:skip                  'timeout occurred, reset for next key

'#########################################################################################################################
'################################################# KEYBOARD 3/Z ##########################################################
'#########################################################################################################################
                        
kb_z                    mov     kbid, #3                'setup keyboard ID
                        mov     f8421, #8               'for key flags: 0001 for kb 0/W, 0010 for kb 1/X, etc.

                        mov     kbcz, ina               'read port and mask clock line
                        and     kbcz, kbcmz

                        cmp     kbcz, #0        wz      'check if clock is low
                        cmp     kbcpz, #1       wc      'check if data bit has already been read
        if_c_and_z      mov     kbcpz, #1               'mark start of low clock pulse (set if clk=z=0 & cp=c=0)
        if_c_and_z      mov     kbdz, ina               'read data line (data is valid 5us before falling edge of clk)
        if_c_and_z      and     kbdz, kbdmz             'mask off data bit
        if_c_and_z      shr     kbdbz, #1               'make way for new bit            
        if_c_and_z      shr     kbdz, #14                'move data bit to msb CHANGE LITERAL (eventually dumps start bit)
        if_c_and_z      or      kbdbz, kbdz             'prepend new bit to data bits (prepending reverses bit order)
        if_c_and_z      add     kbdbcz, #1              'increment the count for the number of received bits (0..11)
        
        if_nc_and_nz    mov     kbcpz, #0               'clock went high so reset clock pulse marker to prepare for next bit

                        cmp     kbdbcz, #11     wc      'check if 11 bits were received (though start bit gets ror'd out)
        if_c_or_z       jmp     #:cont                  'jump if either <11 bits or if clk line is low    

                        and     kbdbz, #$FF             'lop off parity and stop bits

        '############## at this point, a scancode has been received in kbdb* (from a pressed keyboard key)  ##############                

                        cmp     kbdbz, #$AA     wz      'check if keyboard needs reset
        if_z            call    #reset_z
        if_z            mov     delay, d1sec          
        if_z            call    #pause        
        if_z            jmp     #:skip    

        '############## process scan code from keyboard to convert it to ASCII character, including shift   ##############

                        mov     kbdb, kbdbz             'setup for filter call
                        call    #filter                 'preprocess scancode
                        tjnz    jump, #:skip            'jump to skip if jump is set

        '############## at this point, the scancode is not FO, EO nor the byte after those, and also not E1 ##############

                        call    #getchar                'get character and process shift keys (CHANGE?)
                        tjz     rcode, #:skip           'toss out any chars that mapped to 0 that slipped through filter
         
                        'write character to keyboard buffer area of shared hub memory
                        call    #write                  'write character to buffer (if not full)
                        
:skip                   'peform clean up in prep for next byte from keyboard 
                        mov     kbcpz, #0               'reset clock pulse marker
                        mov     kbdbcz, #0              'reset counter for the number of received data bits to zero
                        mov     kbdbz, #0               'clear any residual bits/reset keyboard data bits (REMOVE?)
                        mov     kbtocz, #0              'reset timeout counter                         
                                                        
                        'check for timeout in case of "hang" during a character due to line noise/glitch 
:cont                   tjz     kbdbcz, #end            'check if any of the 11 data bits have been received 
                        add     kbtocz, #1              'if at least 1 bit has been received, increment time-out counter
                        cmp     kbtocz, timeout wc      'compare current count with the timeout constant value
        if_nc           jmp     #:skip                  'timeout occurred, reset for next key

'#########################################################################################################################
                       
end                     jmp     #loop

'#########################################################################################################################                        
'################################################## END MAIN LOOP ######################################################## 
'#########################################################################################################################

                        'could hang if it misses a clock pulse, consider using a timeout
                        'watch out for extended keys causing problems
                        'reset E0/F0 flags after time-out ???

'#########################################################################################################################

write                   'write character (predefined rcode) to circular keyboard buffer area of shared hub memory

                        'get addresses for head and tail pointers for the current keyboard pointer      
                        mov     head, par               'get head ptr for kb buffer
                        add     head, #32               'skip over the 32 bytes of kb buffer area
                        add     head, kbid              'add offset for current kb
                        mov     tail, head              'get tail ptr for kb buffer  
                        add     tail, #4                'add 4-byte offset to point to tail
                        mov     temp, tail              'save a copy of the tail addr for updating later           
                        
                        'read in the head and tail values from the hub, overwriting the addresses
                        rdbyte  head, head              'get the head value (converts addr to value) 
                        rdbyte  tail, tail              'get the tail value (converts addr to value) 

                        'check if buffer is full
                        mov     full, tail              'make a copy of tail
                        add     full, #1                'if (tail + 1)//8 = head, then buffer is full 
                        and     full, #7                'masks off anything > 7, effectively giving modulo 8
                        cmp     full, head      wz      'if equal, then the buffer is full                        
        if_z            jmp     #write_ret              'don't write to buffer since it's full  

                        'write character to buffer
                        mov     buff, par               'get base address of kb buffer area
                        shl     kbid, #3                'convert 0, 1, 2, 3 to 0, 8, 16, 24, respectively
                        add     buff, kbid              'add in offset to point to beginning of current kb buffer
                        add     buff, tail              'add in tail ptr offset (0..7) to point to byte to write  
                        wrbyte  rcode, buff             'write charcter to current kb buffer

                        'adjust tail ptr (seemingly, no locks are needed); note: all heads and tails are 0..7 for 4 kb's
                        add     tail, #1                'increment the tail ptr of the current kb
                        and     tail, #7                'wrap around back to 0 if tail = 8 (mod 8)  
                        wrbyte  tail, temp              'update the tail ptr                         
                        
write_ret               ret
                        
'------------------------------------------------------------------------------------------------------------------------- 

                        'map a code (icode) to another return code (rcode) using a table at a base address (tbase) 
map                     mov     charind, icode          'put code in character index
                        shr     charind, #2             'divide by four, since 4 bytes per 1 long in LUT's
                        add     charind, tbase          'add index into table to base address of lookup table
                        movs    :src, charind           'replace src of retrieval mov downstream
                        nop                             'kill time for the dynamic code modification to complete
:src                    mov     rcode, 0-0              'retrieve long w/char value from LUT (0-0 gets replaced)

                        'isolate the correct byte for the current ascii character in its long
                        mov     charpos, icode          'get the correct byte of long for the current character
                        and     charpos, #%11           'mask all bits but bits 0 and 1; charpos now = 0, 1, 2 or 3
                        shl     charpos, #3             'mult. by 8; gives 0, 8, 16 or 24           
                        shr     rcode, charpos          'shift desired charcter to LSB
                        and     rcode, #$FF             'to handle problem w/ctrl and/or alt being sent (REMOVE LATER?)
map_ret                 ret

'-------------------------------------------------------------------------------------------------------------------------    

                        'preprocess scancodes
                        'handle left/right shift keys ($12/$59), with at least one of them depressed
                        'note: shift keys "repeat" when held down only until another key is pressed
                        'and are then "silent" up until sending their break code upon release
                        'so it's possible for a shift key to be held down and not repeating
                        'and one shift key can stop the repeat of another shift key

filter                  mov     jump, #0                '0 = no jump needed, 1 = jump needed
                        'handle press/release of left shift key
                        cmp     kbdb, #$12      wz      'detect left shift key code
        if_z            or      shift_l, f8421          'flag shift-left...  
        if_z            cmp     kF0, #1         wc      'but if last code was a break indicator...
        if_z_and_nc     muxc    shift_l, f8421          'then clear shift-left flag (if set)... 
        if_z_and_nc     mov     kF0, #0                 'and also clear the break indicator flag        
        if_z            mov     jump, #1                'move on 
                        'handle press/release of right shift key              
                        cmp     kbdb, #$59      wz      'detect right shift key code
        if_z            or      shift_r, f8421          'flag shift-right... 
        if_z            cmp     kF0, #1         wc      'but if last code was a break indicator
        if_z_and_nc     muxc    shift_r, f8421          'then clear shift-right flag (if set)
        if_z_and_nc     mov     kF0, #0                 'and also clear the break indicator flag        
        if_z            mov     jump, #1                'move on
                        'determine if either shift key is pressed and set shift accordingly
                        mov     shift, #0               'clear shift in preparation for OR (to avoid "carry over")                              
                        or      shift, shift_l          'OR's shift bits for all 4 keyboards, not just the current one
                        or      shift, shift_r          'but it doesn't adversely affect the remaining three keyboards

                        'handle special scancodes F0, E0, E1
                        and     kF0, f8421      wz, nr  'check if $F0 (break byte) was issued before the current character
        if_nz           muxz    kF0, f8421              'clear/reset flag for $F0 break byte
        if_nz           mov     jump, #1                'toss out currently received byte (which was preceded by $F0)

                        'remember if F0 was pressed (just now)
                        cmp     kbdb, #$F0      wz      'skip normal break scancode $F0
        if_z            or      kF0, f8421              'flag that the $F0 break code was just received
        if_z            muxnz   kE0, f8421              'clear flag for $E0 break byte, which may or maynot be set        
        if_z            mov     jump, #1                'toss out the received $F0

                        'handle E0, with exceptions made for arrow keys (UDLR scancodes = E0 + 6B/75/72/74)
                        and     kE0, f8421      wc, nr  'check if $E0 (extended byte) was issued before the current character
        if_c            muxz    kE0, f8421              'clear flag for $E0 extended byte
        if_c            cmp     kbdb, #$6B      wz      'check for left-arrow 
        if_c_and_z      jmp     #filter_ret
        if_c            cmp     kbdb, #$75      wz      'check for up-arrow
        if_c_and_z      jmp     #filter_ret
        if_c            cmp     kbdb, #$72      wz      'check for down-arrow
        if_c_and_z      jmp     #filter_ret
        if_c            cmp     kbdb, #$74      wz      'check for right-arrow
        if_c_and_z      jmp     #filter_ret                        
        if_c            mov     jump, #1                'toss out currently received byte (which was preceded by $E0)

                        'remember if E0 was pressed (just now)
                        cmp     kbdb, #$E0      wz      'skip normal extended byte $E0 
        if_z            or      kE0, f8421              'flag that the $E0 extended code was just received
        if_z            mov     jump, #1                'move on                        

                        'handle E1 
                        cmp     kbdb, #$E1      wz      'filter out $E1 used in pause/break
        if_z            mov     jump, #1                'toss out the received $E1

filter_ret              ret      

'-------------------------------------------------------------------------------------------------------------------------

getchar
                        'disregard shift when pressing the space key
                        cmp     kbdb, #$29      wz      'handle space separately, in case a shift key is pressed
        if_z            mov     rcode, #$20             'set character to a space (with or without a shift)
        if_z            jmp     #getchar_ret            'move on

                        'map keyboard scancode to ascii value using map function (in-code:icode, return-code:rcode, tbase)
                        mov     icode, kbdb             'put scancode into icode for map function
                        mov     tbase, #charlut         'setup character lookup table base address (tbase)
                        call    #map                    'map scancode to ascii character; comes back as rcode
                        
                        'throw out any just received scancode that is mapped to zero
                        cmp     rcode, #$00     wz      'ignore $00
        if_z            jmp     #getchar_ret            'move on

                        'check if a left or right shift key is pressed (above code can't handle both at the same time)
                        and     shift, f8421    wz, nr  'check if either the left or right shift key is pressed 
        if_z            jmp     #getchar_ret            'move on since a shift key is not pressed 
                        
                        'handle capital letters (a=$61/97, z=$7A/122, A=$41/65, Z=$5A/90, difference=$20/32)                
                        cmp     rcode, #$61     wc      'check if character code is "a" or greater
        if_c            jmp     #:next                  'character code is lower than "a" so not a letter; move on                      
                        cmp     rcode, #$7B     wc      'check if character code greater than "z"
        if_c            sub     rcode, #$20             'convert lowercase character to uppercase
        if_c            jmp     #getchar_ret            'skip over code for shift of non-alphabetic character

                        'handle shift keys for non-alphabetic keys (there's definitely a shift, but not with a letter key)                      
:next                   cmp     rcode, #$3E     wc      'check if shift occurred within the first shift table ($27~$3D)
        if_c            mov     tbase, #shift1          'setup base address for shift table 1
        if_nc           mov     tbase, #shift2          'use shift table 2 (for values $5B, $5C, $5D and $60 = {|}~)
        if_c            sub     rcode, #$27             'adjust code to coincide with shift1 table, which skips $00~$26
        if_nc           sub     rcode, #$5B             'adjust code to coincide with shift2 table, which skips $00~$5A
                        mov     icode, rcode            'setup input code for map funciton to shift symbol
                        call    #map                    'map non-shifted symbol (icode) to shifted symbol (rcode)
                     
getchar_ret             ret

'------------------------------------------------------------------------------------------------------------------------- 

'%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
'Parallax Code for transmitter, slightly modified
'
' Transmit byte to keyboard
'
transmit
_c1                     or      dira,cmask              'pull clock low
                        mov     delay, d2500            'hold for 125us > 100ux                        
                        call    #pause
_d1                     or      dira,dmask              'pull data low
                        mov     delay, d80              'hold data for ~4us                        
                        call    #pause
_c2                     xor     dira,cmask              'release clock

                        test    data,#$0FF      wc      'append parity and stop bits to byte
                        muxnc   data,#$100
                        or      data,dlsb

                        mov     x,#10                   'ready 10 bits
transmit_bit            call    #wait_c0                'wait until clock low
                        shr     data,#1         wc      'output data bit
_d2                     muxnc   dira,dmask
                        mov     wcond,c1                'wait until clock high
                        call    #wait
                        djnz    x,#transmit_bit         'another bit?

                        mov     wcond,c0d0              'wait until clock and data low
                        call    #wait
                        mov     wcond,c1d1              'wait until clock and data high
                        call    #wait

                        call    #receive_ack            'receive ack byte with timed wait
                        cmp     data,#$FA       wz      'if ack error, reset keyboard
        'if_nz           jmp     #reset

transmit_ret            ret

'
'
' Receive byte from keyboard
'
receive_ack
                        mov     x,#11                   'ready 11 bits
receive_bit             call    #wait_c0                'wait until clock low
                        mov     delay, d320             'pause ~16us                         
                        call    #pause
_d3                     test    dmask,ina       wc      'input data bit
                        rcr     data,#1
                        mov     wcond,c1                'wait until clock high
                        call    #wait
                        djnz    x,#receive_bit          'another bit?

                        shr     data,#22                'align byte
                        test    data,#$1FF      wc      'if parity error, reset keyboard
rand                    and     data,#$FF               'isolate byte

look_ret
receive_ack_ret
receive_ret             ret


'
' Wait for clock/data to be in required state(s)
'
wait_c0                 mov     wcond,c0                '(wait until clock low)

wait                    mov     y,tenms                 'set timeout to 10ms

wloop                   mov     delay, d80              'nap ~4us
                        call    #pause
_c3                     test    cmask,ina       wc      'check required state(s)
_d4                     test    dmask,ina       wz      'loop until got state(s) or timeout
wcond   if_never        djnz    y,#wloop                '(replaced with c0/c1/c0d0/c1d1)

                        'tjz     y,#reset                'if timeout, reset keyboard
wait_ret
wait_c0_ret             ret

                                                     
c0      if_c            djnz    y,#wloop                '(if_never replacements)
c1      if_nc           djnz    y,#wloop
c0d0    if_c_or_nz      djnz    y,#wloop
c1d1    if_nc_or_z      djnz    y,#wloop                

'pause
pause                   djnz    delay, #$               'pause version w/o using cnt
pause_ret               ret

'------------------------------------------------------------------------------------------------------------------------- 

reset_w                 mov     data, #$FF              'Reset kbd W/0 (north)                     
                        mov     dmask, kbdmw           
                        mov     cmask, kbcmw
                        call    #transmit
reset_w_ret             ret

reset_x                 mov     data, #$FF              'Reset kbd X/1 (west)                       
                        mov     dmask, kbdmx             
                        mov     cmask, kbcmx
                        call    #transmit
reset_x_ret             ret

reset_y                 mov     data, #$FF              'Reset kbd Y/2 (south)                        
                        mov     dmask, kbdmy            
                        mov     cmask, kbcmy
                        call    #transmit
reset_y_ret             ret

reset_z                 mov     data, #$FF              'Reset kbd Z/3 (east)                        
                        mov     dmask, kbdmz            
                        mov     cmask, kbcmz
                        call    #transmit
reset_z_ret             ret

'------------------------------------------------------------------------------------------------------------------------- 
'My Vars                                                                                                 
dmask                   long    0                       'data pin mask
cmask                   long    0                       'clock pin mask
dbits                   long    0                       'used for sending data
dbcnt                   long    0                       'count of the number of sent data bits
stbit                   long    1 << 9                  'used to form start bit when transmitting 
timer                   long    $0                      'used for timing
clkst                   long    0                       'desired clk line state (0=low, 1=high)
datst                   long    0                       'desired data line state (0=low, 1=high) 

delay                   long    $0                      'settable delay duration
d1sec                   long    20_000_000              'one second delay
d2500                   long    2500                    '125us/50ns jump branches
d320                    long    320                     ' 16us/50ns jump branches
d80                     long    80                      '  4us/50ns jump branches

kF0                     long    0                       'used to suppress break codes
kE0                     long    0                       'used to suppress extended break codes
shift                   long    0                       'flag for indicating at least one of the shift keys is depressed
shift_l                 long    0                       'flag indicating that the left shift key is currently pressed
shift_r                 long    0                       'flag indicating that the right shift key is currently pressed
f8421                   long    0                       'used to control flag bits for shifts, F0 and E0
timeout                 long    999                     'timeout in case of line error (noise)
                                                        'note: as of last check, slowest kb needed 232 loops
                                                        '232 loops / 11 bits = ~21 loops per bit (interesting)

shiftw                  long    0                       'flag for indicating at least one of the shift keys is depressed
shiftx                  long    0
shifty                  long    0
shiftz                  long    0

lshiftw                 long    0                       'flag indicating that the left shift key is currently pressed 
lshiftx                 long    0
lshifty                 long    0
lshiftz                 long    0

rshiftw                 long    0                       'flag indicating that the right shift key is currently pressed
rshiftx                 long    0
rshifty                 long    0
rshiftz                 long    0

icode                   long    0                       'input code to map from one form to another (for map function)
rcode                   long    0                       'return/result code after being mapped (for map function)
tbase                   long    0                       'base address of table to use when mapping (for map function)

                                                        'Note: Only one kb handler accesses vars in this section at a time
kbid                    long    0                       'keyboard ID number
buff                    long    0                       'key buffer
head                    long    0                       'head of key buffer
tail                    long    0                       'tail of key buffer
full                    long    0                       'used to check if key buffer is full
temp                    long    0                       'gen. purpose, such as for copying tail addr
dbg1                    long    1                       'debug 1 = 1

kbcw                    long    0                       'Keyboard Clock Lines
kbcx                    long    0
kbcy                    long    0
kbcz                    long    0
                                                                                
kbdw                    long    0                       'Keyboard Data lines 
kbdx                    long    0
kbdy                    long    0
kbdz                    long    0

kbdbw                   long    0                       'keyboard Data Bits (progressively built)
kbdbx                   long    0
kbdby                   long    0
kbdbz                   long    0
kbdb                    long    0                       'kb data bits w,x,y,z for filter function
jump                    long    0                       'flags whether jump to skip is needed

kbcpw                   long    0                       'marks start of low portion of clk pulse
kbcpx                   long    0
kbcpy                   long    0
kbcpz                   long    0

kbdbcw                  long    0                       'count of the number of received data bits
kbdbcx                  long    0
kbdbcy                  long    0
kbdbcz                  long    0

kbtocw                  long    0                       'counter to implement a time out in case of line glitch (noise)
kbtocx                  long    0
kbtocy                  long    0
kbtocz                  long    0

                                'KBD Data-clock pairs: 7-8, 12-13, 16-17, 23-24 (after change of P15/16 to P16/17)
kbdmw                   long    %0000_0000_0000_0000_0000_0000_1000_0000        'Keyboard Data Masks   
kbdmx                   long    %0000_0000_0000_0000_0001_0000_0000_0000
kbdmy                   long    %0000_0000_0000_0001_0000_0000_0000_0000
kbdmz                   long    %0000_0000_1000_0000_0000_0000_0000_0000
   
kbcmw                   long    %0000_0000_0000_0000_0000_0001_0000_0000        'Keyboard Clock Masks 
kbcmx                   long    %0000_0000_0000_0000_0010_0000_0000_0000
kbcmy                   long    %0000_0000_0000_0010_0000_0000_0000_0000
kbcmz                   long    %0000_0001_0000_0000_0000_0000_0000_0000

'--------------------------------------------------------------------------------------------------------------------------

charind                 long    0
charpos                 long    0                       'indicates which byte of long to use for charlut

                                       'SC              Key          
charlut                 byte    $00    '00              stored little endian w/LSB first
                        byte    $D8    '01              F9
                        byte    $00    '02
                        byte    $D4    '03              F5
                        byte    $D2    '04              F3
                        byte    $D0    '05              F1
                        byte    $D1    '06              F2
                        byte    $DB    '07              F12
                        byte    $00    '08                  
                        byte    $D9    '09              F10
                        byte    $D7    '0A              F8
                        byte    $D5    '0B              F6
                        byte    $D3    '0C              F4
                        byte    $09    '0D              Tab
                        byte    $60    '0E              `        accute accent "`"                        
                        byte    $00    '0F
                        byte    $00    '10
                        byte    $00    '11    Alt-R     Alt-L    ZERO'D
                        byte    $F0    '12              Shift-L  (12 is also sent in prtscn but is ignored due to E0/F0)
                        byte    $00    '13
                        byte    $00    '14    Ctrl-R    Ctrl-L   ZERO'D  (also sent in break key 8-byte sequence)                
                        byte    $71    '15              q
                        byte    $31    '16              1
                        byte    $00    '17
                        byte    $00    '18
                        byte    $00    '19
                        byte    $7A    '1A              z
                        byte    $73    '1B              s
                        byte    $61    '1C              a    <-------------------------- a = scancode 1C
                        byte    $77    '1D              w
                        byte    $32    '1E              2
                        byte    $00    '1F    Win-L            ZERO'D
                        byte    $00    '20
                        byte    $63    '21              c
                        byte    $78    '22              x
                        byte    $64    '23              d
                        byte    $65    '24              e
                        byte    $34    '25              4
                        byte    $33    '26              3
                        byte    $00    '27    Win-R
                        byte    $00    '28
                        byte    $20    '29              Space
                        byte    $76    '2A              v
                        byte    $66    '2B              f
                        byte    $74    '2C              t
                        byte    $72    '2D              r
                        byte    $35    '2E              5
                        byte    $00    '2F    Apps
                        byte    $00    '30
                        byte    $6E    '31              n
                        byte    $62    '32              b
                        byte    $68    '33              h
                        byte    $67    '34              g
                        byte    $79    '35              y
                        byte    $36    '36              6
                        byte    $00    '37    Pwr
                        byte    $00    '38
                        byte    $00    '39
                        byte    $6D    '3A              m
                        byte    $6A    '3B              j
                        byte    $75    '3C              u
                        byte    $37    '3D              7
                        byte    $38    '3E              8
                        byte    $00    '3F    Sleep
                        byte    $00    '40
                        byte    $2C    '41              ,
                        byte    $6B    '42              k
                        byte    $69    '43              i
                        byte    $6F    '44              o
                        byte    $30    '45              0
                        byte    $39    '46              9
                        byte    $00    '47
                        byte    $00    '48
                        byte    $2E    '49              .
                        byte    $2F    '4A    (/)       /
                        byte    $6C    '4B              l
                        byte    $3B    '4C              ;
                        byte    $70    '4D              p
                        byte    $2D    '4E              -
                        byte    $00    '4F               
                        byte    $00    '50
                        byte    $00    '51
                        byte    $27    '52              '        apostrophe
                        byte    $00    '53
                        byte    $5B    '54              [
                        byte    $3D    '55              =
                        byte    $00    '56
                        byte    $00    '57
                        byte    $DE    '58              CapsLock
                        byte    $F1    '59              Shift-R
                        byte    $0D    '5A    (Enter)   Enter
                        byte    $5D    '5B              ]
                        byte    $00    '5C
                        byte    $5C    '5D              \
                        byte    $00    '5E    WakeUp
                        byte    $00    '5F
                        byte    $00    '60
                        byte    $00    '61
                        byte    $00    '62
                        byte    $00    '63
                        byte    $00    '64
                        byte    $00    '65
                        byte    $C8    '66              BackSpace
                        byte    $00    '67
                        byte    $00    '68
                        byte    $00    '69    End       (1)      ZERO'D
                        byte    $00    '6A
                        byte    $E4    '6B    Lt-arrow  (4)
                        byte    $00    '6C    Home      (7)      ZERO'D
                        byte    $00    '6D
                        byte    $00    '6E
                        byte    $00    '6F
                        byte    $00    '70    Ins       (0)      ZERO'D
                        byte    $00    '71    Del       (.)      ZERO'D
                        byte    $E2    '72    Dn-arrow  (2)
                        byte    $00    '73              (5)      ZERO'D                 
                        byte    $E6    '74    Rt-arrow  (6)
                        byte    $E8    '75    Up-arrow  (8)      (LUDR = E4,E8,E2,E6 ~ 4826, E+6B,75,72,74)
                        byte    $CB    '76              Esc      ($CB/$1B?)   <------ CHECK
                        byte    $00    '77              NumLock  ZERO'D (also sent in break key 8-byte sequence)              
                        byte    $DA    '78              F11
                        byte    $00    '79              (+)      ZERO'D
                        byte    $00    '7A    PgDn      (3)      ZERO'D
                        byte    $00    '7B              (-)      ZERO'D
                        byte    $00    '7C    PringScr  (*)      ZERO'D (also sent in print-screen sequence)
                        byte    $00    '7D    PgUp      (9)      ZERO'D
                        byte    $00    '7E              ScrLock  ZERO'D
                        byte    $00    '7F      
                        byte    $00    '80                          
                        byte    $00    '81                                            
                        byte    $00    '82              
                        byte    $D6    '83              F7  {note: table fills out a long so next table is on a long boundary}
                                                
                        'Non-alphabetic key shifts are divided into two mapping tables: shift1 and shift2...
                        'based on ASCII values of the unshifted keys mapped to ASCII values for shifted keys
                        'They are mapped this way because the span of the unshifted ASCII values (i.e., the 
                        'range from highest value to lowest) was less than the span/range for the scancodes.
                        'The two tables only waste 8 bytes for unmapped/unused bytes (the zeros in the tables)
shift1                  byte    $22, $00, $00, $00, $00, "<_>?)!@#$%^&*(", $00, ":", $00, "+", $00  {last byte pads long}
shift2                  byte    "{|}", $00, $00, "~" '5B->7B, 5C->3A, 5D->7D, SKIP, SKIP, 60->7E {table begins on a long}                     

'--------------------------------------------------------------------------------------------------------------------------
'Parallax Vars
dlsb                    long    1 << 9        
tenms                   long    10_000 / 4               

data                    res     1
t                       res     1
x                       res     1
y                       res     1

'-------------------------------------------------------------------------------------------------------------------------

                        fit     444             'last known use was 444 longs 
                        